@@ -19,14 +19,16 @@ from stubber.utils.my_version import __version__
     is_eager=True,
 )
 @click.pass_context
-def stubber_cli(ctx, verbose: int = 0):
+def stubber_cli(ctx: click.Context, verbose: int = 0) -> None:
     # ensure that ctx.obj exists and is a dict (in case `cli()` is called
     ctx.ensure_object(dict)
     # replace std log handler with a custom one capped on INFO level
     log.remove()
     level = {0: "INFO", 1: "DEBUG", 2: "TRACE"}.get(verbose, "TRACE")
     if level == "INFO":
-        format_str = "<green>{time:YY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{module: <18}</cyan> - <level>{message}</level>"
+        format_str = (
+            "<green>{time:YY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <cyan>{module: <18}</cyan> - <level>{message}</level>"
+        )
     else:
         format_str = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"
 
