@@ -1,24 +1,27 @@
 """
- Processing for createstubs.py
- minimizes and cross-compiles a micropyton file.
+Processing for createstubs.py
+Minimizes and cross-compiles a MicroPyton file.
 """
-import io
 import itertools
 import subprocess
 import tempfile
 from pathlib import Path
-from typing import List, Tuple, Union, IO
+from typing import List, Tuple, Union
 from contextlib import ExitStack
+from io import BytesIO, StringIO, IOBase, TextIOWrapper
 
 from loguru import logger as log
 
-try:
-    import python_minifier
-except ImportError:  # pragma: no cover
-    python_minifier = None
+import python_minifier
+
+# Type Aliases for minify
+StubSource = Union[Path, str, StringIO, TextIOWrapper]
+XCompileDest = Union[Path, BytesIO]
+LineEdits = List[Tuple[str, str]]
 
 
-def edit_lines(content, edits, diff=False):
+def edit_lines(content: str, edits: LineEdits, diff: bool = False):
+    # sourcery skip: no-long-functions
     """Edit string by list of edits
 
     Args:
@@ -36,29 +39,32 @@ def edit_lines(content, edits, diff=False):
         str: edited string
     """
 
-    def comment(l, x):
+    def comment(l: str, x: str):
+        """Comment out line, so it will be removed on minify"""
         return l.replace(x, f"# {x}")
 
-    def rprint(l, x):  # type: ignore # lgtm [py/unused-local-variable] pylint: disable= unused-variable
+    def rprint(l: str, x: str):  # type: ignore
+        """Replace (logging) with print"""
         split = l.split("(")
         if len(split) > 1:
             return l.replace(split[0].strip(), "print")
         return l.replace(x, "print")
 
-    def rpass(l, x):  # type: ignore # lgtm [py/unused-local-variable] pylint: disable= unused-variable
+    def rpass(l: str, x: str):  # type: ignore
+        """Replace with pass"""
         return l.replace(x, "pass")
 
-    def get_whitespace_context(content, index):
+    def get_whitespace_context(content: List[str], index: int):
         """Get whitespace count of lines surrounding index"""
 
-        def count_ws(line):
+        def count_ws(line: str):
             return sum(1 for _ in itertools.takewhile(str.isspace, line))
 
-        lines = content[index - 1 : index + 2]
+        lines = content[index - 1 : index + 2]  # BUG - index out of range results in too few results returned to caller
         context = (count_ws(l) for l in lines)
         return context
 
-    def handle_multiline(content, index):
+    def handle_multiline(content: List[str], index: int):
         """Handles edits that require multiline comments
 
         Example:
@@ -105,7 +111,7 @@ def edit_lines(content, edits, diff=False):
         if check and line_ws != post_ws:
             return range(index - 1, index + 1)
 
-    def handle_try_except(content, index):
+    def handle_try_except(content: List[str], index: int):
         """Checks if line at index is in try/except block
 
         Handles situations like this:
@@ -125,20 +131,20 @@ def edit_lines(content, edits, diff=False):
 
     lines = []
     multilines = set()
-    content = content.splitlines(keepends=True)
-    for line in content:
+    content_l = content.splitlines(keepends=True)
+    for line in content_l:
         _line = line
         for edit, text in edits:
             if text in line:
                 if edit == "comment":
-                    l_index = content.index(line)
+                    l_index = content_l.index(line)
                     # Check if edit spans multiple lines
-                    mline = handle_multiline(content, l_index)
+                    mline = handle_multiline(content_l, l_index)
                     if mline:
                         multilines.update(mline)
                         break
                     # Check if line is only statement in try/except
-                    if handle_try_except(content, l_index):
+                    if handle_try_except(content_l, l_index):
                         edit = "rpass"
                         text = line.strip()
                 func = eval(edit)  # pylint: disable= eval-used
@@ -157,19 +163,37 @@ def edit_lines(content, edits, diff=False):
     return stripped
 
 
-def minify_script(source_script: Union[Path, str, IO[str]], keep_report=True, diff=False) -> str:
-    """minifies createstubs.py
+def minify_script(source_script: StubSource, keep_report: bool = True, diff: bool = False) -> str:
+    """
+    Minifies createstubs.py and variants
 
     Args:
-        keep_report (bool, optional): Keeps single report line in createstubs
+    source_script:
+        - (str): content to edit
+        - (Path): path to file to edit
+        - (IOBase): file-like object to edit
+    keep_report (bool, optional): Keeps single report line in createstubs
             Defaults to True.
-        diff (bool, optional): Print diff from edits. Defaults to False.
+    diff (bool, optional): Print diff from edits. Defaults to False.
 
     Returns:
         str: minified source text
     """
 
-    edits: List[Tuple[str, str]] = [
+    source_content = ""
+    if isinstance(source_script, Path):
+        source_content = source_script.read_text()
+    elif isinstance(source_script, (StringIO, TextIOWrapper)):
+        source_content = "".join(source_script.readlines())
+    elif isinstance(source_script, str):  # type: ignore
+        source_content = source_script
+    else:
+        raise TypeError(f"source_script must be str, Path, or file-like object, not {type(source_script)}")
+
+    if not source_content:
+        raise ValueError("No source content")
+
+    edits: LineEdits = [
         ("comment", "print"),
         ("comment", "import logging"),
         # report keepers may be inserted here
@@ -189,6 +213,7 @@ def minify_script(source_script: Union[Path, str, IO[str]], keep_report=True, di
         ("comment", "_log.warning"),
     ]
     if keep_report:
+        # insert report keepers after the comment modifiers
         edits[2:2] = [
             # keepers
             ("rprint", 'self._log.info("Stub module: '),
@@ -197,18 +222,6 @@ def minify_script(source_script: Union[Path, str, IO[str]], keep_report=True, di
             ("rprint", 'self._log.info("Created stubs for'),
         ]
 
-    if not python_minifier:  # pragma: no cover
-        raise ModuleNotFoundError("python_minifier not available")
-
-    source_content = source_script
-    if isinstance(source_script, str) and Path(source_script).exists():
-        source_script = Path(source_script)
-        source_content = source_script.read_text()
-    elif isinstance(source_script, Path):
-        source_content = source_script.read_text()
-    else:
-        source_content = "".join(source_script.readlines())
-
     content = edit_lines(source_content, edits, diff=diff)
 
     source = python_minifier.minify(
@@ -221,7 +234,7 @@ def minify_script(source_script: Union[Path, str, IO[str]], keep_report=True, di
         rename_locals=True,  # short names save memory
         preserve_locals=["stubber", "path"],  # names to keep
         rename_globals=True,  # short names save memory
-        # keep these globals to allow testing/mocking to work against the minified version
+        # keep these globals to allow testing/mocking to work against the minified not compiled version
         preserve_globals=[
             "main",
             "Stubber",
@@ -242,57 +255,97 @@ def minify_script(source_script: Union[Path, str, IO[str]], keep_report=True, di
 
 
 def minify(
-    source: Union[str, Path, IO[str]],
-    target: Union[str, Path, IO[str], IO[bytes]],
+    source: StubSource,
+    target: StubSource,
     keep_report: bool = True,
     diff: bool = False,
-    cross_compile: bool = False,
 ):
+    """Minifies and compiles a script"""
+    source_buf = None
+    target_buf = None
+
     with ExitStack() as stack:
-        source_buf = source
-        target_buf = target
 
-        if isinstance(source, (str, Path)):
-            source = Path(source)
+        if isinstance(source, Path):
             source_buf = stack.enter_context(source.open("r"))
-            if isinstance(target, (str, Path)):
-                target = Path(target)
-                # if target is a folder , then append the filename
-                if target.exists() and target.is_dir():
-                    target = target / Path(source).name
-                target_buf = stack.enter_context(target.open("w+"))
+        elif isinstance(source, (StringIO, str)):
+            # different types of file-like objects are both acepted by minify_script
+            source_buf = source
+        else:
+            raise TypeError(f"source must be str, Path, or file-like object, not {type(source)}")
 
+        if isinstance(target, Path):
+            if target.is_dir():
+                if isinstance(source, Path):
+                    target = target / source.name
+                else:
+                    target = target / "minified.py"  # or raise error?
+            target_buf = stack.enter_context(target.open("w+"))
+        elif isinstance(target, IOBase):  # type: ignore
+            target_buf = target
         try:
             minified = minify_script(source_script=source_buf, keep_report=keep_report, diff=diff)
-            if isinstance(target_buf, (io.StringIO, io.TextIOWrapper)):
-                target_buf.write(minified)
+            target_buf.write(minified)
         except Exception as e:  # pragma: no cover
             log.exception(e)
-        else:
-            return _extracted_from_minify_(target, cross_compile, target_buf, minified)
+    return 0
 
 
-# TODO Rename this here and in `minify`
-def _extracted_from_minify_(target, cross_compile, target_buf, minified):
-    log.debug("Minified file written to :", target)
-    if not cross_compile:
-        return 0
-    cross_targ = target
-    if not isinstance(cross_targ, Path):
-        _, temp_file = tempfile.mkstemp()
-        temp_file = Path(temp_file)
-        target_buf.seek(0)
-        temp_file.write_text(minified)
-        cross_targ = temp_file
-    result = subprocess.run(["mpy-cross", "-O2", str(cross_targ)])
+def cross_compile(
+    source: StubSource,
+    target: XCompileDest,
+    version: str = "",
+):  # sourcery skip: assign-if-exp
+    """Runs mpy-cross on a (minified) script"""
+    # Sources can be a file, a string, or a file-like object
+    if isinstance(source, Path):
+        source_file = source
+    elif isinstance(source, str):
+        # create a temp file and write the source to it
+        source_file = write_to_temp_file(source)
+    elif isinstance(source, StringIO):
+        source_file = write_to_temp_file(source.getvalue())
+    else:
+        raise TypeError(f"source must be str, Path, or file-like object, not {type(source)}")
+
+    _target = None
+    if isinstance(target, Path):
+        if target.is_dir():
+            target = target / source.name if isinstance(source, Path) else target / "minified.mpy"
+        _target = target.with_suffix(".mpy")
+    else:
+        # target must be a Path object
+        _target = get_temp_file(suffix=".mpy")
+
+    cmd = ["pipx", "run", f"mpy-cross=={version}"] if version else ["pipx", "run", "mpy-cross"]
+    # Add params
+    cmd += ["-O2", str(source_file), "-o", str(_target), "-s", "createstubs.py"]
+    log.trace(" ".join(cmd))
+    result = subprocess.run(cmd, capture_output=True, text=True)
     if result.returncode == 0:
-        if isinstance(target_buf, io.BytesIO):
-            target_buf.write(cross_targ.read_bytes())
-        else:
-            mpy_target = (
-                target.with_suffix(".mpy")
-                if hasattr(target, "with_suffix")
-                else target
-            )
-            log.debug("mpy-cross compiled to    :", mpy_target)
+        log.debug(f"mpy-cross compiled to    : {_target.name}")
+    else:
+        log.error("mpy-cross failed to compile:")
+
+    if isinstance(target, BytesIO):
+        # copy the byte contents of the temp file to the target file-like object
+        with _target.open("rb") as f:
+            target.write(f.read())
+        # _target.unlink()
+
     return result.returncode
+
+
+def write_to_temp_file(source: str):
+    """Writes a string to a temp file and returns the Path object"""
+    _, temp_file = tempfile.mkstemp(suffix=".py", prefix="mpy_cross_")
+    temp_file = Path(temp_file)
+    temp_file.write_text(source)
+    return temp_file
+
+
+def get_temp_file(prefix: str = "mpy_cross_", suffix: str = ".py"):
+    """Get temp file and returns the Path object"""
+    _, temp_file = tempfile.mkstemp(prefix=prefix, suffix=suffix)
+    temp_file = Path(temp_file)
+    return temp_file
