@@ -1,3 +1,8 @@
+"""
+Read versions published to PyPi or test.PyPi.
+uses the pypi-simple package to get the versions from the simple index.
+"""
+
 from typing import Optional
 
 from packaging.version import Version, parse
@@ -5,7 +10,7 @@ from pypi_simple import PyPISimple, NoSuchProjectError
 from loguru import logger as log
 
 
-def get_pypy_versions(package_name: str, base: Optional[Version] = None, production: bool = True):
+def get_pypi_versions(package_name: str, base: Optional[Version] = None, production: bool = True):
     """Get all versions of a package from a PyPI endpoint."""
     package_info = None
     if production:
@@ -16,7 +21,7 @@ def get_pypy_versions(package_name: str, base: Optional[Version] = None, product
     try:
         with PyPISimple(endpoint=endpoint) as client:
             package_info = client.get_project_page(project=package_name)
-    except NoSuchProjectError as e:
+    except NoSuchProjectError:
         log.debug(f"Package {package_name} not found on {endpoint}")
         return []
 
