@@ -89,14 +89,17 @@ class ManifestMetadata:
         self.version = None
         self.description = None
         self.license = None
+        self.author = None
 
-    def update(self, description=None, version=None, license=None):
+    def update(self, description=None, version=None, license=None, author=None):
         if description:
             self.description = description
         if version:
             self.version = version
         if license:
             self.license = version
+        if author:
+            self.author = author
 
 
 # Turns a dict of options into a object with attributes used to turn the
@@ -167,6 +170,7 @@ class ManifestFile:
             "freeze_as_str": self.freeze_as_str,
             "freeze_as_mpy": self.freeze_as_mpy,
             "freeze_mpy": self.freeze_mpy,
+            # "convert_path": self._resolve_path,  # compatibility for old manifest files
             "options": IncludeOptions(**kwargs),
         }
 
@@ -210,7 +214,11 @@ class ManifestFile:
                 raise ManifestFileError("Expected .py file")
             kind = KIND_COMPILE_AS_MPY
 
-        self._manifest_files.append(ManifestOutput(FILE_TYPE_LOCAL, full_path, target_path, timestamp, kind, self._metadata[-1], opt))
+        self._manifest_files.append(
+            ManifestOutput(
+                FILE_TYPE_LOCAL, full_path, target_path, timestamp, kind, self._metadata[-1], opt
+            )
+        )
 
     def _search(self, base_path, package_path, files, exts, kind, opt=None, strict=False):
         base_path = self._resolve_path(base_path)
@@ -244,7 +252,7 @@ class ManifestFile:
             if base_path:
                 os.chdir(prev_cwd)  # type: ignore
 
-    def metadata(self, description=None, version=None, license=None):
+    def metadata(self, description=None, version=None, license=None, author=None):
         """
         From within a manifest file, use this to set the metadata for the
         package described by current manifest.
@@ -253,7 +261,7 @@ class ManifestFile:
         to obtain the metadata for the top-level manifest file.
         """
 
-        self._metadata[-1].update(description, version, license)
+        self._metadata[-1].update(description, version, license, author)
         return self._metadata[-1]
 
     def include(self, manifest_path, top_level=False, **kwargs):
@@ -303,7 +311,9 @@ class ManifestFile:
                 try:
                     exec(f.read(), self._manifest_globals(kwargs))
                 except Exception as er:
-                    raise ManifestFileError("Error in manifest file: {}: {}".format(manifest_path, er))
+                    raise ManifestFileError(
+                        "Error in manifest file: {}: {}".format(manifest_path, er)
+                    )
                 os.chdir(prev_cwd)
             if not top_level:
                 self._metadata.pop()
@@ -322,12 +332,14 @@ class ManifestFile:
                 lib_dirs = ["unix-ffi"] + lib_dirs
 
             for lib_dir in lib_dirs:
-                for manifest_path in glob.glob(
-                    os.path.join(self._path_vars["MPY_LIB_DIR"], lib_dir, "**", name, "manifest.py"),
-                    recursive=True,
+                # Search for {lib_dir}/**/{name}/manifest.py.
+                for root, dirnames, filenames in os.walk(
+                    os.path.join(self._path_vars["MPY_LIB_DIR"], lib_dir)
                 ):
-                    self.include(manifest_path, **kwargs)
-                    return
+                    if os.path.basename(root) == name and "manifest.py" in filenames:
+                        self.include(root, **kwargs)
+                        return
+
             raise ValueError("Library not found in local micropython-lib: {}".format(name))
         else:
             # TODO: HTTP request to obtain URLs from manifest.json.
@@ -406,28 +418,37 @@ class ManifestFile:
         `opt` is the optimisation level to pass to mpy-cross when compiling .py
         to .mpy.
         """
-        self._freeze_internal(path, script, exts=(".py", ".mpy"), kind=KIND_FREEZE_AUTO, opt=opt)
+        self._freeze_internal(
+            path,
+            script,
+            exts=(
+                ".py",
+                ".mpy",
+            ),
+            kind=KIND_FREEZE_AUTO,
+            opt=opt,
+        )
 
     def freeze_as_str(self, path):
         """
         Freeze the given `path` and all .py scripts within it as a string,
         which will be compiled upon import.
         """
-        self._search(path, None, None, exts=(".py"), kind=KIND_FREEZE_AS_STR)
+        self._search(path, None, None, exts=(".py",), kind=KIND_FREEZE_AS_STR)
 
     def freeze_as_mpy(self, path, script=None, opt=None):
         """
         Freeze the input (see above) by first compiling the .py scripts to
         .mpy files, then freezing the resulting .mpy files.
         """
-        self._freeze_internal(path, script, exts=(".py"), kind=KIND_FREEZE_AS_MPY, opt=opt)
+        self._freeze_internal(path, script, exts=(".py",), kind=KIND_FREEZE_AS_MPY, opt=opt)
 
     def freeze_mpy(self, path, script=None, opt=None):
         """
         Freeze the input (see above), which must be .mpy files that are
         frozen directly.
         """
-        self._freeze_internal(path, script, exts=(".mpy"), kind=KIND_FREEZE_MPY, opt=opt)
+        self._freeze_internal(path, script, exts=(".mpy",), kind=KIND_FREEZE_MPY, opt=opt)
 
 
 # Generate a temporary file with a line appended to the end that adds __version__.
